module hylltac.

accum_sig hyllgoal.
accumulate listmanip.

kind	  answer	type.

type	  yes		answer.

type	init		goal -> goal -> o.
type	copy		int -> goal -> goal -> o.
type	andM_r		list int -> goal -> goal -> o.
type	andM_l		int -> goal -> goal -> o.
type	imp_r		goal -> goal -> o.
type	imp_l		int -> list int -> goal -> goal -> o.
type	andA_r		goal -> goal -> o.
type	andA_l1		int -> goal -> goal -> o.
type	andA_l2		int -> goal -> goal -> o.
type	orA_r1		goal -> goal -> o.
type	orA_r2		goal -> goal -> o.
type	orA_l		int -> goal -> goal -> o.
type	wall_l		int -> int -> goal -> goal -> o.
type	wall_l1		int -> (int -> int) -> int -> goal -> goal -> o.
type	bang_l		int -> goal -> goal -> o.
type	at_r		goal -> goal -> o.
type	at_l		int -> goal -> goal -> o.
type	down_r		goal -> goal -> o.
type	down_l		int -> goal -> goal -> o.
type	dagger_l	int -> goal -> goal -> o.
type	delta_r		goal -> goal -> o.
type	step_r		goal -> goal -> o.
type	step_l		int -> goal -> goal -> o.
type	iunfold_l	int -> goal -> goal -> o.
type	lunfold_l	int -> goal -> goal -> o.

%type    exists_i_tac            goal -> goal -> o.
%type    forall_e_tac            int -> goal -> goal -> o.
%type    exists_e_tac            int -> goal -> goal -> o.
%type    exists_e_name_tac       int -> string -> goal -> goal -> o.
%type    exists_e_tacr           int -> goal -> goal -> o.
%type    exists_e_name_tacr      int -> string -> goal -> goal -> o.

local   reduce_world    int -> int -> o.
local   remove_zero     int -> int -> o.
type    simplify_world  int -> int -> o.
local   same_world      int -> int -> o.

reduce_world (U + N1 + N2) (V + M1) :- !,
  reduce_world U (V + M), M1 is (M + N1 + N2).
reduce_world (U + N) (U + M) :- M is N, !.
reduce_world U (U + 0).

remove_zero (W + 0) W :- !.
remove_zero W W.

simplify_world U V :- reduce_world U W, remove_zero W V.

same_world W W :- !.
same_world U V :- simplify_world U U1, simplify_world V V1, !, U1 = V1.

init (s_init # (seq Sig G ((A @ W1)::nil) (A @ W2))) truegoal :-
  same_world W1 W2.

copy N ((s_copy J' (J'::D) Prf)  # (seq Sig G D J))
       (Prf # (seq Sig G (J'::D) J)) :-
  nth_item N J' G.

andM_r Ns1 ((s_ConjMR D1 Prf1 D2 Prf2) # (seq Sig G D (A *o B @ W)))
           (andgoal (Prf1 # (seq Sig G D1 (A @ W)))
	     	    (Prf2 # (seq Sig G D2 (B @ W)))) :-
  divide_list Ns1 D D1 D2.

andM_l N ((s_ConjML A B U D2 D' Prf) # (seq Sig G D J))
         (Prf # (seq Sig G D' J)) :-
  nth_item_replace N (A *o B @ U) ((A @ U)::(B @ U)::nil) D D',
  nth_item_and_rest N (A @ U) D' D1,
  nth_item_and_rest N (B @ U) D1 D2.

imp_r ((s_ImpR D' Prf) # (seq Sig G D (A ->> B @ W)))
      (Prf # (seq Sig G D' (B @ W))) :-
  D' = ((A @ W)::D).

imp_l N Ns1 ((s_ImpL A B U DRest1 DRest2 ((B @ U)::DRest2) Prf1 Prf2) #
	     (seq Sig G D J))
            (andgoal (Prf1 # (seq Sig G DRest1 (A @ U)))
                     (Prf2 # (seq Sig G ((B @ U)::DRest2) J))) :-
  nth_item_divide_rest N Ns1 (A ->> B @ U) D DRest1 DRest2.

andA_r ((s_ConjAR Prf1 Prf2) # (seq Sig G D (A &a B @ W)))
       (andgoal (Prf1 # (seq Sig G D (A @ W)))
    		(Prf2 # (seq Sig G D (B @ W)))).

andA_l1 N ((s_ConjAL1 A B U DRest D' Prf) # (seq Sig G D J))
          (Prf # (seq Sig G D' J)) :-
  nth_item_replace N (A &a B @ U) ((A @ U)::nil) D D',
  nth_item_and_rest N (A &a B @ U) D DRest.

andA_l2 N ((s_ConjAL2 A B U DRest D' Prf) # (seq Sig G D J))
          (Prf # (seq Sig G D' J)) :-
  nth_item_replace N (A &a B @ U) ((B @ U)::nil) D D',
  nth_item_and_rest N (A &a B @ U) D DRest.

orA_r1 ((s_DisjAR1 Prf) # (seq Sig G D (A +o B @ W)))
       (Prf # (seq Sig G D (A @ W))).

orA_r2 ((s_DisjAR2 Prf) # (seq Sig G D (A +o B @ W)))
       (Prf # (seq Sig G D (B @ W))).

orA_l N ((s_DisjAL A B U DRest D1 D2 Prf1 Prf2) # (seq Sig G D J))
        (andgoal (Prf1 # (seq Sig G D1 J))
		 (Prf2 # (seq Sig G D2 J))) :-
  nth_item_replace N (A +o B @ U) ((A @ U)::nil) D D1,
  nth_item_replace N (A +o B @ U) ((B @ U)::nil) D D2,
  nth_item_and_rest N (A +o B @ U) D DRest.

wall_l N V ((s_WallL A U V DRest D' Prf) # (seq Sig G D J))
	   (Prf # (seq Sig G D' J)) :-
  nth_item_replace N (wall A @ U) (((A V) @ U)::nil) D D',
  nth_item_and_rest N (wall A @ U) D DRest.

wall_l1 N E M ((s_WallL A U V DRest D' Prf) # (seq Sig G D J))
	      (Prf # (seq Sig G D' J)) :-
  nth_item M (sig_w W) Sig,
  V = (E W),
  nth_item_replace N (wall A @ U) (((A V) @ U)::nil) D D',
  nth_item_and_rest N (wall A @ U) D DRest.

bang_l N ((s_BangL A U D' Prf) # (seq Sig G D J))
	 (Prf # (seq Sig ((A @ U)::G) D' J)) :-
  nth_item_and_rest N (!! A @ U) D D'.

at_r ((s_AtR Prf) # (seq Sig G D ((A at U) @ W)))
     (Prf # (seq Sig G D (A @ U))).

at_l N ((s_AtL A U V DRest D' Prf) # (seq Sig G D J))
       (Prf # (seq Sig G D' J)) :-
  nth_item_replace N ((A at U) @ V) ((A @ U)::nil) D D',
  nth_item_and_rest N ((A at U) @ V) D DRest.

down_r ((s_DownR Prf) # (seq Sig G D ((down A) @ W)))
       (Prf # (seq Sig G D ((A W) @ W))).

down_l N ((s_DownL A V DRest D' Prf) # (seq Sig G D J))
	 (Prf # (seq Sig G D' J)) :-
  nth_item_replace N (down A @ V) (((A V) @ V)::nil) D D',
  nth_item_and_rest N (down A @ V) D DRest.

dagger_l N ((s_unfold_l "dagger" Prf) # (seq Sig G D J))
	   (Prf # (seq Sig G D' J)) :-
  nth_item_replace N (dagger A @ U) (((wall u\ (A at u)) @ U)::nil) D D'.

delta_r ((s_unfold_r "delt_" (A' @ W) Prf) # (seq Sig G D (delta V A @ W)))
        (Prf # (seq Sig G D (A' @ W))) :-
  A' = (down (u\ (A at (u + V)))).

step_r ((s_stepR Prf) # (seq Sig G D (step A @ W)))
       (Prf # (seq Sig G D (A' @ W))) :-
  A' = (down (w\ (A at (w + 1)))).

step_l N ((s_unfold_l "step" Prf) # (seq Sig G D J))
	 (Prf # (seq Sig G D' J)) :-
  nth_item_replace N (step A @ U) ((A' @ U)::nil) D D',
  A' = (down (u\ (A at (u + 1)))).

iunfold_l N ((d_unfold Def Prf) # (seq Sig G D J))
            (Prf # (seq Sig G' D J)) :-
  nth_item_replace N (fdef Def J') ((J' A)::nil) G G',
  Def A.

lunfold_l N ((d_unfold Def Prf) # (seq Sig G D J))
            (Prf # (seq Sig G D' J)) :-
  nth_item_replace N (fdef Def J') ((J' A)::nil) D D',
  Def A.
